--[[=============================================================================
    Get, Handle and Dispatch message functions

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_messages = "2015.03.31"
end

--[[=============================================================================
    GetMessage()
  
    Description:
    Used to retrieve a message from the communication buffer. Each driver is
    responsible for parsing that communication from the buffer.
  
    Parameters:
    None
  
    Returns:
    A single message from the communication buffer
===============================================================================]]
function GetMessage()
	local message, pos
	
	--TODO: Implement a string using Lua captures and patterns which 
	--		will be used by string.match to parse out a single message
	--      from the receive buffer(gReceiveBuffer).
	--		The example shown here will return all characters from the beginning of 
	--		the gReceiveBuffer up until but not including the first carriage return.
	local pattern = "!(.-)\r()"
	
	if (gReceiveBuffer:len() > 0) then
		message, pos = string.match(gReceiveBuffer, pattern)
		if (message == nil) then
			--LOG:Info("Do not have a complete message")
			return ""
		end
		gReceiveBuffer = gReceiveBuffer:sub(pos)		
	end

	return message
	
end

--[[=============================================================================
    HandleMessage(message)]

    Description
    This is where we parse the messages returned from the GetMessage()
    function into a command and data. The call to 'DispatchMessage' will use the
    'name' variable as a key to determine which handler routine, function, should
    be called in the DEV_MSG table. The 'value' variable will then be passed as
    a string parameter to that routine.

    Parameters
    message(string) - Message string containing the function and value to be sent to
                      DispatchMessage

    Returns
    Nothing
===============================================================================]]
function HandleMessage(message)
	if (type(message) == "table") then
		LogTrace("HandleMessage. Message is ==>")
		LogTrace(message)
		LogTrace("<==")
	else
		LogTrace("HandleMessage. Message is ==>%s<==", message)
	end

	local pattern =  ""

	if (message:find("PONG") ~= nil) then
	   return
	elseif (message:find("AUDIN") ~= nil) then
	   pattern = "(%w+)%((%d+)%)"
	elseif (message:find("AUDMODE") ~= nil) then
	   pattern = "(%w+)%((%d+)%)(.*)"
	elseif (message:find("AUDTYPE") ~= nil) then
	   pattern = "(%w+)%((.+)%)"
	elseif (message:find("ZMUTE") ~= nil) then	--this line needs to be before find("MUTE") line...
	   pattern = "(ZMUTE)(%a+)"	   
	elseif (message:find("MUTE") ~= nil) then
	   pattern = "(MUTE)(%a+)"
	elseif (message:find("POWER") ~= nil) then
	   pattern = "(%w+)%((%d+)%)"
	elseif (message:find("RPFOC") ~= nil) then
	   pattern = "(%w+)%((%d+)%)(.*)"
	elseif (message:find("RPVOI") ~= nil) then
	   pattern = "(%w+)%((%d+)%)(.*)"
	elseif (message:find("SRC") ~= nil) then
	   pattern = "(%w+)%((%d+)%)(.*)"
	elseif (message:find("TRIMBASS") ~= nil) then
	   pattern = "(%w+)%((.+)%)"
	elseif (message:find("TRIMTREB") ~= nil) then
	   pattern = "(%w+)%((.+)%)"	   
	elseif (message:find("VIDIN") ~= nil) then
	   pattern = "(%w+)%((%d+)%)(.*)"
	elseif (message:find("VIDTYPE") ~= nil) then
	   pattern = "(%w+)%((.+)%)"
	elseif (message:find("VOL") ~= nil) then
	   --pattern = "(%w+)%((%d+)%)"
	   pattern = "(%a+)%((.+)%)"
	elseif (message:find("POWERZONE2") ~= nil) then
	   pattern = "(%w+)%((%d+)%)"
	elseif (message:find("ZAUDIN") ~= nil) then
	   pattern = "(%w+)%((%d+)%)"
	elseif (message:find("ZSRC") ~= nil) then
	   pattern = "(%w+)%((%d+)%)(.*)"
	elseif (message:find("ZVOL") ~= nil) then
	   --pattern = "(%w+)%((%d+)%)"
	   pattern = "(%a+)%((.+)%)"
	else
	   --pattern = "(.+)"
	   pattern = "(%w+)%((%d+)%)"
	end
	
	local name, value, value2 = string.match(message, pattern)
     --print(message, name, value, tostring(value2))
	name = name or message
	value = value or ""	

	DispatchMessage(name, value, value2)
	
end

--[[=============================================================================
    DispatchMessage(MsgKey, MsgData)

    Description
    Parse routine that will call the routines to handle the information returned
    by the connected system.

    Parameters
    MsgKey(string)  - The function to be called from within DispatchMessage
    MsgData(string) - The parameters to be passed to the function found in MsgKey

    Returns
    Nothing
===============================================================================]]
function DispatchMessage(MsgKey, MsgData, MsgData2)
	if (DEV_MSG[MsgKey] ~= nil and (type(DEV_MSG[MsgKey]) == "function")) then
		LogInfo("DEV_MSG." .. tostring(MsgKey) .. ":  " .. tostring(MsgData))
		local status, err = pcall(DEV_MSG[MsgKey], MsgData, MsgData2)
		if (not status) then
			LogError("LUA_ERROR: " .. err)
		end
	else
		LogTrace("HandleMessage: Unhanded command = " .. MsgKey)
	end
end

--[[
TODO: Create DEV_MSG functions for all messages to call Notifies.
	  Sample functions are included below for all applicable notifications.
--]]

function DEV_MSG.AUDIN(value)
	LogTrace("DEV_MSG.AUDIN(),  Main Audio Input = " .. tAudioInputResponseMap[value])

end

function DEV_MSG.VIDIN(value)
	LogTrace("DEV_MSG.VIDIN(), Main Video Input = " .. tVideoInputResponseMap[value])
end

function DEV_MSG.ZAUDIN(value)
	LogTrace("DEV_MSG.ZAUDIN(),  Zone B Audio Input = " .. tAudioInputResponseMap[value])

end

function DEV_MSG.AUDTYPE(value)
    LogTrace("DEV_MSG.AUDTYPE: " .. value)
end

function DEV_MSG.VIDTYPE(value)
    LogTrace("DEV_MSG.VIDTYPE: " .. value)
end

function DEV_MSG.STREAMTYPE(value)
    LogTrace("DEV_MSG.STREAMTYPE(), value = " .. value)
end

function DEV_MSG.ZSTREAMTYPE(value)
    LogTrace("DEV_MSG.ZSTREAMTYPE(), value = " .. value)
end

gAudioModeCount = 0
function DEV_MSG.AUDMODECOUNT(value)
    LogTrace("DEV_MSG.AUDMODECOUNT(), value = " .. value)
    gAudioModeCount = tonumber(value)
    tAudioModeMap = {}
end

function DEV_MSG.AUDMODE(value, value2)
    LogTrace("DEV_MSG.AUDMODE(), value = " .. value .. ", name = " .. value2)
    local index = string.gsub(value2, '"', '')
    tAudioModeMap[index] = value
end

gFocusPositionCount = 0
function DEV_MSG.RPFOCCOUNT(value)
    LogTrace("DEV_MSG.RPFOCCOUNT(), value = " .. value)
    gFocusPositionCount = tonumber(value)
    tFocusPositionMap = {}
end

function DEV_MSG.RPFOC(value, value2)
    LogTrace("DEV_MSG.RPFOC(), value = " .. value .. ", name = " .. value2)
    local index = string.gsub(value2, '"', '')
    tFocusPositionMap[index] = value
end

gVoicingCount = 0
function DEV_MSG.RPVOICOUNT(value)
    LogTrace("DEV_MSG.RPFOCCOUNT(), value = " .. value)
    gVoicingCount = tonumber(value)
    tVoicingMap = {}
end

function DEV_MSG.RPVOI(value, value2)
    LogTrace("DEV_MSG.RPVOI(), value = " .. value .. ", name = " .. value2)
    local index = string.gsub(value2, '"', '')
    tVoicingMap[index] = value
end

function DEV_MSG.SRC(value, value2)
	LogTrace("DEV_MSG.SRC(), value = " .. value .. ", name = " .. value2)
	local input = tInputConnMapByName[tInputResponseMap[value]].ID 	--mod 1000 value of Input Connection ID
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output = 0 												--mod 1000 value of Output Connection ID		
	
     local bindingID = tInputConnMapByID[input].BindingID 			--both Receiver and Tuner proxies support input selection	
	LogInfo("INPUT_OUTPUT_CHANGED, input = " .. tInputResponseMap[value] .. ", output = " .. tOutputConnMap[output])
	gReceiverProxy:dev_InputOutputChanged(bindingID, input, output)
end

function DEV_MSG.ZSRC(value, value2)
     if (value == "0") then LogInfo("DEV_MSG.ZSRC(), Zone B is Following Main.") return end
	
	LogTrace("DEV_MSG.ZSRC(), value = " .. value .. ", name = " .. value2)
	local input = tInputConnMapByName[tZoneBInputResponseMap[value]].ID 	--mod 1000 value of Input Connection ID
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output = 1 												--mod 1000 value of Output Connection ID		
	
     local bindingID = tInputConnMapByID[input].BindingID 			--both Receiver and Tuner proxies support input selection	
	LogInfo("INPUT_OUTPUT_CHANGED, input = " .. tInputResponseMap[value] .. ", output = " .. tOutputConnMap[output])
	gReceiverProxy:dev_InputOutputChanged(bindingID, input, output)
end

function DEV_MSG.POWER(value)
	LogTrace("DEV_MSG.POWER(), value = " .. value)
	
	local output = 0 												--mod 1000 value of Output Connection ID
	if (value == "1") then
		gReceiverProxy:dev_PowerOn(output)
	elseif (value== "0") then
		gReceiverProxy:dev_PowerOff(output)
	else
		LogWarn("DEV_MSG.POWER(): value not valid - " .. value)
	end		
end

function DEV_MSG.POWERZONE2(value)
	LogTrace("DEV_MSG.POWERZONE2(), value = " .. value)
	
	local output = 1 												--mod 1000 value of Output Connection ID
	if (value == "1") then
		gReceiverProxy:dev_PowerOn(output)
	elseif (value== "0") then
		gReceiverProxy:dev_PowerOff(output)
	else
		LogWarn("DEV_MSG.POWERZONE2(): value not valid - " .. value)
	end		
end

function DEV_MSG.VOL(value)
	LogTrace("DEV_MSG.VOL(), value = " .. value)
	
	local output = 0 --mod 1000 value of Output Connection ID
	local deviceLevel = value
	local minDeviceLevel = -999
	local maxDeviceLevel = 240
	--C4 volume level uses a percentage scale: 0 - 100
	local c4Level = ConvertVolumeToC4(deviceLevel, minDeviceLevel, maxDeviceLevel)	
	--local c4Level = value
	
	gReceiverProxy:dev_VolumeLevelChanged(output, c4Level, deviceLevel)		
end

function DEV_MSG.ZVOL(value)
	LogTrace("DEV_MSG.ZVOL(), value = " .. value)
	
	local output = 1 --mod 1000 value of Output Connection ID
	local deviceLevel = value
	local minDeviceLevel = -999
	local maxDeviceLevel = 240
	--C4 volume level uses a percentage scale: 0 - 100
	local c4Level = ConvertVolumeToC4(deviceLevel, minDeviceLevel, maxDeviceLevel)	
	--local c4Level = value
	
	gReceiverProxy:dev_VolumeLevelChanged(output, c4Level, deviceLevel)		
end

function DEV_MSG.MUTE(value)
	LogTrace("DEV_MSG.MUTE(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output = 0    --mod 1000 value of Output Connection ID
	
	local state   	--Mute state represented as "True" or "False"
	-- TODO: values "01" & "00" will need to be modified based upon the device protocol specification
	if (value == "ON") then 
		state = "True"
	elseif (value == "OFF") then
		state = "False"
	else
		LogWarn("DEV_MSG.MUTE(), value not valid, exiting...")
		return
	end
	gReceiverProxy:dev_MuteChanged(output, state)
end	

function DEV_MSG.ZMUTE(value)
	LogTrace("DEV_MSG.ZMUTE(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output = 1    --mod 1000 value of Output Connection ID
	
	local state   	--Mute state represented as "True" or "False"
	-- TODO: values "01" & "00" will need to be modified based upon the device protocol specification
	if (value == "ON") then 
		state = "True"
	elseif (value == "OFF") then
		state = "False"
	else
		LogWarn("DEV_MSG.MUTE(), value not valid, exiting...")
		return
	end
	gReceiverProxy:dev_MuteChanged(output, state)
end	

function DEV_MSG.TRIMBASS(value)
	LogTrace("DEV_MSG.BASS(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output = 0  --mod 1000 value of Output Connection ID
	
	-- TODO: set "level", Bass level is represented as a percentage value
	local level = value
	
	--gReceiverProxy:dev_BassLevelChanged(output, level)
end	

function DEV_MSG.TRIMTREB(value)
	LogTrace("DEV_MSG.TREB(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output = 0 --mod 1000 value of Output Connection ID
	
	-- TODO: set "level", Treble level is represented as a percentage value
	local level = value
	
	--gReceiverProxy:dev_TrebleLevelChanged(output, level)
end	

function DEV_MSG.BALANCE(value)
	LogTrace("DEV_MSG.BALANCE(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output  --mod 1000 value of Output Connection ID
	
	-- TODO: set "level", Bass level is represented as a percentage value
	local level  
	
	gReceiverProxy:dev_BalanceLevelChanged(output, level)
end	

function DEV_MSG.LOUDNESS(value)
	LogTrace("DEV_MSG.LOUDNESS(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output  --mod 1000 value of Output Connection ID
	
	-- TODO: set "state", Loudness state is represented as "True" or "False" (literal string, not boolean)
	local state  
	gReceiverProxy:dev_LoudnessChanged(output, state)
end

function DEV_MSG.SURROUND(value)
	-- TODO: ensure that tSurroundResponses table is indexed correctly so that "value" will return the correct mode_id
	local mode_id = tSurroundResponses[value].ID
	
	-- TODO: set "output" to mod 1000 value of Output Connection ID
	local output = 0
	gReceiverProxy:dev_SurroundModeChanged(output, mode_id)
end

--AM/FM Tuner Status
function DEV_MSG.TUNER(value)
	LogTrace("DEV_MSG.TUNER(), value = " .. value)
	local channel = tonumber_loc(value)
	local input
	if channel > 1760 then
		input = tInputConnMapByName['FM Antenna'].ID + 3000 --3000 Audio, 1000 Video
	else
		input = tInputConnMapByName['AM Antenna'].ID + 3000 --3000 Audio, 1000 Video
	end		
	gReceiverProxy:dev_TunerStatus(input, channel)
end
	
